/*
 * Copyright (c) 1999-2002 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * "Portions Copyright (c) 1999-2002 Apple Computer, Inc.  All Rights
 * Reserved.  This file contains Original Code and/or Modifications of
 * Original Code as defined in and that are subject to the Apple Public
 * Source License Version 1.0 (the 'License').  You may not use this file
 * except in compliance with the License.  Please obtain a copy of the
 * License at http://www.apple.com/publicsource and read it before using
 * this file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License."
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
	File:		makehfs.c

	Contains:	Initialization code for HFS and HFS Plus volumes.

	Copyright:	 1984-1999 by Apple Computer, Inc., all rights reserved.

*/

#include <unistd.h>
#include <err.h>
#include <fcntl.h>

#include <sys/param.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <pwd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if __APPLE__
#include <CoreFoundation/CFString.h>
extern Boolean _CFStringGetFileSystemRepresentation(CFStringRef string, UInt8 *buffer, CFIndex maxBufLen);
#endif


#include <hfs/hfs_format.h>
#include "hfs_endian.h"

#include "newfs_hfs.h"
#include "readme.h"


#define HFS_BOOT_DATA	"/usr/share/misc/hfsbootdata"

#define HFS_JOURNAL_FILE	".journal"
#define HFS_JOURNAL_INFO	".journal_info_block"

#define kJournalFileType	0x6a726e6c	/* 'jrnl' */


typedef HFSMasterDirectoryBlock HFS_MDB;

struct filefork {
	UInt16	startBlock;
	UInt16	blockCount;
	SInt32	logicalSize;
	SInt32	physicalSize;
};

struct filefork	gDTDBFork, gSystemFork, gReadMeFork;


static void WriteMDB __P((const DriveInfo *driveInfo, HFS_MDB *mdbp));
static void InitMDB __P((hfsparams_t *defaults, UInt32 driveBlocks, HFS_MDB *mdbp));

static void WriteVH __P((const DriveInfo *driveInfo, HFSPlusVolumeHeader *hp));
static void InitVH __P((hfsparams_t *defaults, UInt64 sectors,
		HFSPlusVolumeHeader *header));

static void WriteBitmap __P((const DriveInfo *dip, UInt32 startingSector,
        UInt32 alBlksUsed, UInt8 *buffer));

static void WriteExtentsFile __P((const DriveInfo *dip, UInt32 startingSector,
        const hfsparams_t *dp, HFSExtentDescriptor *bbextp, void *buffer,
        UInt32 *bytesUsed, UInt32 *mapNodes));
static void InitExtentsRoot __P((UInt16 btNodeSize, HFSExtentDescriptor *bbextp,
		void *buffer));

static void WriteCatalogFile __P((const DriveInfo *dip, UInt32 startingSector,
        const hfsparams_t *dp, HFSPlusVolumeHeader *header, void *buffer,
        UInt32 *bytesUsed, UInt32 *mapNodes));
static void WriteJournalInfo(const DriveInfo *driveInfo, UInt32 startingSector,
			     const hfsparams_t *dp, HFSPlusVolumeHeader *header,
			     void *buffer);
static void InitCatalogRoot_HFSPlus __P((const hfsparams_t *dp, const HFSPlusVolumeHeader *header, void * buffer));
static void InitCatalogRoot_HFS __P((const hfsparams_t *dp, void * buffer));
static void InitFirstCatalogLeaf __P((const hfsparams_t *dp, void * buffer,
		int wrapper));
static void InitSecondCatalogLeaf __P((const hfsparams_t *dp, void *buffer));

static void WriteDesktopDB(const hfsparams_t *dp, const DriveInfo *driveInfo,
        UInt32 startingSector, void *buffer, UInt32 *mapNodes);

static void ClearDisk __P((const DriveInfo *driveInfo, UInt64 startingSector,
		UInt32 numberOfSectors));
static void WriteSystemFile __P((const DriveInfo *dip, UInt32 startingSector,
		UInt32 *filesize));
static void WriteReadMeFile __P((const DriveInfo *dip, UInt32 startingSector,
		UInt32 *filesize));
static void WriteMapNodes __P((const DriveInfo *driveInfo, UInt32 diskStart,
		UInt32 firstMapNode, UInt32 mapNodes, UInt16 btNodeSize, void *buffer));
static void WriteBuffer __P((const DriveInfo *driveInfo, UInt64 startingSector,
		UInt32 byteCount, const void *buffer));
static UInt32 Largest __P((UInt32 a, UInt32 b, UInt32 c, UInt32 d ));

static void MarkBitInAllocationBuffer __P((HFSPlusVolumeHeader *header,
		UInt32 allocationBlock, void* sectorBuffer, UInt32 *sector));

static UInt32 GetDefaultEncoding();

static UInt32 UTCToLocal __P((UInt32 utcTime));

static UInt32 DivideAndRoundUp __P((UInt32 numerator, UInt32 denominator));

static int ConvertUTF8toUnicode __P((const UInt8* source, UInt32 bufsize,
		UniChar* unibuf, u_big_int16_t *charcount));


void SETOFFSET (void *buffer, UInt16 btNodeSize, SInt16 recOffset, SInt16 vecOffset);
#define SETOFFSET(buf,ndsiz,offset,rec)		\
	SET_SWAP_BE16(*(u_big_int16_t*)((UInt8 *)(buf) + (ndsiz) + (-2 * (rec))), (offset))

#define BYTESTOBLKS(bytes,blks)		DivideAndRoundUp((bytes),(blks))

#define ROUNDUP(x, u)	(((x) % (u) == 0) ? (x) : ((x)/(u) + 1) * (u))

/*
 * make_hfs
 *	
 * This routine writes an initial HFS volume structure onto a volume.
 * It is assumed that the disk has already been formatted and verified.
 * 
 * For information on the HFS volume format see "Data Organization on Volumes"
 * in "Inside Macintosh: Files" (p. 2-52).
 * 
 */
int
make_hfs(const DriveInfo *driveInfo,
	  hfsparams_t *defaults,
	  UInt32 *plusSectors,
	  UInt32 *plusOffset)
{
	UInt32 sector;
	UInt32 diskBlocksUsed;
	UInt32 mapNodes;
	UInt32 sectorsPerBlock;
	void *nodeBuffer = NULL;
	HFS_MDB	*mdbp = NULL;
	UInt32 bytesUsed;
	
	*plusSectors = 0;
	*plusOffset = 0;

	/* assume sectorSize <= blockSize */
	sectorsPerBlock = defaults->blockSize / driveInfo->sectorSize;
	

	/*--- CREATE A MASTER DIRECTORY BLOCK:  */

	mdbp = (HFS_MDB*)malloc((size_t)kBytesPerSector);
	nodeBuffer = malloc(8192);  /* max bitmap bytes is 8192 bytes */
	if (nodeBuffer == NULL || mdbp == NULL) 
		exit(1);

	/* MDB Initialized in native byte order */
	InitMDB(defaults, driveInfo->totalSectors, mdbp);

	
	/*--- ZERO OUT BEGINNING OF DISK (bitmap and b-trees):  */
	
	diskBlocksUsed = (SWAP_BE16(mdbp->drAlBlSt) + 1) +
		(SWAP_BE32(mdbp->drXTFlSize) + SWAP_BE32(mdbp->drCTFlSize)) / kBytesPerSector;
	if (defaults->flags & kMakeHFSWrapper) {
		diskBlocksUsed += MAX(kDTDB_Size, SWAP_BE32(mdbp->drAlBlkSiz)) / kBytesPerSector;
		diskBlocksUsed += MAX(sizeof(hfswrap_readme), SWAP_BE32(mdbp->drAlBlkSiz)) / kBytesPerSector;
		diskBlocksUsed += MAX(24 * 1024, SWAP_BE32(mdbp->drAlBlkSiz)) / kBytesPerSector;
	}
	ClearDisk(driveInfo, 0, diskBlocksUsed);
	/* also clear out last 8 sectors (4K) */
	ClearDisk(driveInfo, driveInfo->totalSectors - 8, 8);

	/* If this is a wrapper, add boot files... */
	if (defaults->flags & kMakeHFSWrapper) {

		sector = SWAP_BE16(mdbp->drAlBlSt) + 
			 SWAP_BE32(mdbp->drXTFlSize)/kBytesPerSector +
			 SWAP_BE32(mdbp->drCTFlSize)/kBytesPerSector;

		WriteDesktopDB(defaults, driveInfo, sector, nodeBuffer, &mapNodes);

		if (mapNodes > 0)
			WriteMapNodes(driveInfo, (sector + 1), 1, mapNodes, kHFSNodeSize, nodeBuffer);

		gDTDBFork.logicalSize = MAX(kDTDB_Size, SWAP_BE32(mdbp->drAlBlkSiz));
		gDTDBFork.startBlock = (sector - SWAP_BE16(mdbp->drAlBlSt)) / sectorsPerBlock;
		gDTDBFork.blockCount = BYTESTOBLKS(gDTDBFork.logicalSize, SWAP_BE32(mdbp->drAlBlkSiz));
		gDTDBFork.physicalSize = gDTDBFork.blockCount * SWAP_BE32(mdbp->drAlBlkSiz);

		sector += gDTDBFork.physicalSize / kBytesPerSector;

		WriteReadMeFile(driveInfo, sector, &gReadMeFork.logicalSize);
		gReadMeFork.startBlock = gDTDBFork.startBlock + gDTDBFork.blockCount;
		gReadMeFork.blockCount = BYTESTOBLKS(gReadMeFork.logicalSize, SWAP_BE32(mdbp->drAlBlkSiz));
		gReadMeFork.physicalSize = gReadMeFork.blockCount * SWAP_BE32(mdbp->drAlBlkSiz);

		sector += gReadMeFork.physicalSize / kBytesPerSector;

		WriteSystemFile(driveInfo, sector, &gSystemFork.logicalSize);
		gSystemFork.startBlock = gReadMeFork.startBlock + gReadMeFork.blockCount;
		gSystemFork.blockCount = BYTESTOBLKS(gSystemFork.logicalSize, SWAP_BE32(mdbp->drAlBlkSiz));
		gSystemFork.physicalSize = gSystemFork.blockCount * SWAP_BE32(mdbp->drAlBlkSiz);

		SET_SWAP_BE16(mdbp->drFreeBks, SWAP_BE16(mdbp->drFreeBks) - gDTDBFork.blockCount +
				   	gReadMeFork.blockCount +
				   	gSystemFork.blockCount +
					defaults->hfsWrapperFreeBlks);
		SET_SWAP_BE16(mdbp->drEmbedExtent.startBlock, SWAP_BE16(mdbp->drNmAlBlks) - SWAP_BE16(mdbp->drFreeBks));
		SET_SWAP_BE16(mdbp->drEmbedExtent.blockCount, SWAP_BE16(mdbp->drFreeBks));
		SET_SWAP_BE16(mdbp->drFreeBks, defaults->hfsWrapperFreeBlks);
	}


	/*--- WRITE ALLOCATION BITMAP TO DISK:  */

	WriteBitmap(driveInfo, SWAP_BE16(mdbp->drVBMSt), SWAP_BE16(mdbp->drNmAlBlks) - SWAP_BE16(mdbp->drFreeBks), nodeBuffer);


	/*--- WRITE FILE EXTENTS B*-TREE TO DISK:  */

	sector = SWAP_BE16(mdbp->drAlBlSt);	/* reset */
	WriteExtentsFile(driveInfo, sector, defaults, &mdbp->drEmbedExtent, nodeBuffer, &bytesUsed, &mapNodes);

	if (mapNodes > 0)
		WriteMapNodes(driveInfo, (sector + bytesUsed/kBytesPerSector),
			bytesUsed/kHFSNodeSize, mapNodes, kHFSNodeSize, nodeBuffer);

	sector += (SWAP_BE32(mdbp->drXTFlSize)/kBytesPerSector);


	/*--- WRITE CATALOG B*-TREE TO DISK:  */

	WriteCatalogFile(driveInfo, sector, defaults, NULL, nodeBuffer, &bytesUsed, &mapNodes);

	if (mapNodes > 0)
		WriteMapNodes(driveInfo, (sector + bytesUsed/kBytesPerSector),
			bytesUsed/kHFSNodeSize, mapNodes, kHFSNodeSize, nodeBuffer);


	/*--- WRITE MASTER DIRECTORY BLOCK TO DISK:  */
	
	*plusSectors = SWAP_BE16(mdbp->drEmbedExtent.blockCount) *
		(SWAP_BE32(mdbp->drAlBlkSiz) / driveInfo->sectorSize);
	*plusOffset = SWAP_BE16(mdbp->drAlBlSt) + SWAP_BE16(mdbp->drEmbedExtent.startBlock) *
		(SWAP_BE32(mdbp->drAlBlkSiz) / driveInfo->sectorSize);

	/* write mdb last in case we fail along the way */

	/* Writes both copies of the MDB */
	WriteMDB (driveInfo, mdbp);
	/* MDB is now big-endian */

	free(nodeBuffer);		
	free(mdbp);	

	return (0);
}


/*
 * make_hfs
 *	
 * This routine writes an initial HFS volume structure onto a volume.
 * It is assumed that the disk has already been formatted and verified.
 *
 */
int
make_hfsplus(const DriveInfo *driveInfo, hfsparams_t *defaults)
{
	UInt16			btNodeSize;
	UInt32			sector;
	UInt32			sectorsPerBlock;
	UInt32			volumeBlocksUsed;
	UInt32			mapNodes;
	UInt32			sectorsPerNode;
	void			*nodeBuffer = NULL;
	HFSPlusVolumeHeader	*header = NULL;
	UInt32			temp;
	UInt32			bits;
	UInt32 bytesUsed;

	/* --- Create an HFS Plus header:  */

	header = (HFSPlusVolumeHeader*)malloc((size_t)kBytesPerSector);
	if (header == NULL)
		exit(1);

	/* VH Initialized in native byte order */
	InitVH(defaults, driveInfo->totalSectors, header);

	sectorsPerBlock = SWAP_BE32(header->blockSize) / kBytesPerSector;


	/*--- ZERO OUT BEGINNING OF DISK:  */

	volumeBlocksUsed = SWAP_BE32(header->totalBlocks) - SWAP_BE32(header->freeBlocks) - 1;
	if ( SWAP_BE32(header->blockSize) == 512 )
		volumeBlocksUsed--;
	bytesUsed = (SWAP_BE32(header->totalBlocks) - SWAP_BE32(header->freeBlocks)) * sectorsPerBlock;
	ClearDisk(driveInfo, 0, bytesUsed);
	/* also clear out last 8 sectors (4K) */
	ClearDisk(driveInfo, driveInfo->totalSectors - 8, 8);

	/*--- Allocate a buffer for the rest of our IO:  */

	temp = Largest( defaults->catalogNodeSize * 2,
			defaults->extentsNodeSize,
			SWAP_BE32(header->blockSize),
			(SWAP_BE32(header->totalBlocks) - SWAP_BE32(header->freeBlocks)) / 8 );
	/* 
	 * If size is not a mutiple of 512, round up to nearest sector
	 */
	if ( (temp & 0x01FF) != 0 )
		temp = (temp + kBytesPerSector) & 0xFFFFFE00;
	
	nodeBuffer = valloc((size_t)temp);
	if (nodeBuffer == NULL)
		exit(1);


		
	/*--- WRITE ALLOCATION BITMAP BITS TO DISK:  */

	sector = SWAP_BE32(header->allocationFile.extents[0].startBlock) * sectorsPerBlock;
	bits = SWAP_BE32(header->totalBlocks) - SWAP_BE32(header->freeBlocks);
	bits -= (SWAP_BE32(header->blockSize) == 512) ? 2 : 1;
	WriteBitmap(driveInfo, sector, bits, nodeBuffer);

	/*
	 * Write alternate Volume Header bitmap bit to allocations file at
	 * 2nd to last sector on HFS+ volume
	 */
	if (SWAP_BE32(header->totalBlocks) > kBitsPerSector)
		bzero(nodeBuffer, kBytesPerSector);
	MarkBitInAllocationBuffer( header, SWAP_BE32(header->totalBlocks) - 1, nodeBuffer, &sector );

	if ( SWAP_BE32(header->blockSize) == 512 ) {
		UInt32	sector2;
		MarkBitInAllocationBuffer( header, SWAP_BE32(header->totalBlocks) - 2,
			nodeBuffer, &sector2 );
		
		/* cover the case when altVH and last block are on different bitmap sectors. */
		if ( sector2 != sector ) {
			bzero(nodeBuffer, kBytesPerSector);
			MarkBitInAllocationBuffer(header, SWAP_BE32(header->totalBlocks) - 1,
				nodeBuffer, &sector);
			WriteBuffer(driveInfo, sector, kBytesPerSector, nodeBuffer);

			bzero(nodeBuffer, kBytesPerSector);
			MarkBitInAllocationBuffer(header, SWAP_BE32(header->totalBlocks) - 2,
				nodeBuffer, &sector);
		}
	}
	WriteBuffer(driveInfo, sector, kBytesPerSector, nodeBuffer);


	/*--- WRITE FILE EXTENTS B-TREE TO DISK:  */

	btNodeSize = defaults->extentsNodeSize;
	sectorsPerNode = btNodeSize/kBytesPerSector;

	sector = SWAP_BE32(header->extentsFile.extents[0].startBlock) * sectorsPerBlock;
	WriteExtentsFile(driveInfo, sector, defaults, NULL, nodeBuffer, &bytesUsed, &mapNodes);

	if (mapNodes > 0)
		WriteMapNodes(driveInfo, (sector + bytesUsed/kBytesPerSector),
			bytesUsed/btNodeSize, mapNodes, btNodeSize, nodeBuffer);

	
	/*--- WRITE CATALOG B-TREE TO DISK:  */
	
	btNodeSize = defaults->catalogNodeSize;
	sectorsPerNode = btNodeSize/kBytesPerSector;

	sector = SWAP_BE32(header->catalogFile.extents[0].startBlock) * sectorsPerBlock;
	WriteCatalogFile(driveInfo, sector, defaults, header, nodeBuffer, &bytesUsed, &mapNodes);

	if (mapNodes > 0)
		WriteMapNodes(driveInfo, (sector + bytesUsed/kBytesPerSector),
			bytesUsed/btNodeSize, mapNodes, btNodeSize, nodeBuffer);

	/*--- JOURNALING SETUP */
	if (defaults->journaledHFS) {
	    sector = SWAP_BE32(header->journalInfoBlock) * sectorsPerBlock;
	    WriteJournalInfo(driveInfo, sector, defaults, header, nodeBuffer);
	}


	/*--- WRITE VOLUME HEADER TO DISK:  */

	/* write header last in case we fail along the way */

	/* Writes both copies of the volume header */
	WriteVH (driveInfo, header);
	/* VH is now big-endian */

	free(nodeBuffer);
	free(header);

	return (0);
}


/*
 * WriteMDB
 *
 * Writes the Master Directory Block (MDB) to disk.
 *
 * The MDB is byte-swapped if necessary to big endian. Since this
 * is always the last operation, there's no point in unswapping it.
 */
static void
WriteMDB (const DriveInfo *driveInfo, HFS_MDB *mdbp)
{
	//SWAP_HFSMDB (mdbp);

	WriteBuffer(driveInfo, kMDBStart, kBytesPerSector, mdbp);
	WriteBuffer(driveInfo, driveInfo->totalSectors - 2, kBytesPerSector, mdbp);
}


/*
 * InitMDB
 *
 * Initialize a Master Directory Block (MDB) record.
 * 
 * If the alignment parameter is non-zero, it indicates the aligment
 * (in 512 byte sectors) that should be used for allocation blocks.
 * For example, if alignment is 8, then allocation blocks will begin
 * on a 4K boundary relative to the start of the partition.
 *
 */
static void
InitMDB(hfsparams_t *defaults, UInt32 driveBlocks, HFS_MDB *mdbp)
{
	UInt32	alBlkSize;
	UInt16	numAlBlks;
	UInt32	timeStamp;
	UInt16	bitmapBlocks;
	UInt32	alignment;

	alignment = defaults->hfsAlignment;
	bzero(mdbp, kBytesPerSector);
	
	alBlkSize = defaults->blockSize;

	/* calculate the number of sectors needed for bitmap (rounded up) */
	if (defaults->flags & kMakeMaxHFSBitmap)
		bitmapBlocks = kHFSMaxAllocationBlks / kBitsPerSector;
	else
		bitmapBlocks = ((driveBlocks / (alBlkSize >> kLog2SectorSize)) +
				kBitsPerSector-1) / kBitsPerSector;

	SET_SWAP_BE16(mdbp->drAlBlSt, kVolBitMapStart + bitmapBlocks);  /* in sectors (disk blocks) */

	/* If requested, round up block start to a multiple of "alignment" blocks */
	if (alignment != 0)
		SET_SWAP_BE16(mdbp->drAlBlSt, ((SWAP_BE16(mdbp->drAlBlSt) + alignment - 1) / alignment) * alignment);
	
	/* Now find out how many whole allocation blocks remain... */
	numAlBlks = (driveBlocks - SWAP_BE16(mdbp->drAlBlSt) - kTailBlocks) /
			(alBlkSize >> kLog2SectorSize);

	timeStamp = UTCToLocal(defaults->createDate);
	
	SET_SWAP_BE16(mdbp->drSigWord, kHFSSigWord);
	SET_SWAP_BE32(mdbp->drCrDate, timeStamp);
	SET_SWAP_BE32(mdbp->drLsMod, timeStamp);
	SET_SWAP_BE16(mdbp->drAtrb, kHFSVolumeUnmountedMask);
	SET_SWAP_BE16(mdbp->drVBMSt, kVolBitMapStart);
	SET_SWAP_BE16(mdbp->drNmAlBlks, numAlBlks);
	SET_SWAP_BE32(mdbp->drAlBlkSiz, alBlkSize);
	SET_SWAP_BE32(mdbp->drClpSiz, defaults->dataClumpSize);
	SET_SWAP_BE32(mdbp->drNxtCNID, defaults->nextFreeFileID);
	SET_SWAP_BE16(mdbp->drFreeBks, numAlBlks);

#if __APPLE__
	/*
	 * Map UTF-8 input into a Mac encoding.
	 * On conversion errors "untitled" is used as a fallback.
	 */
	{
		UniChar unibuf[kHFSMaxVolumeNameChars];
		CFStringRef cfstr;
		CFIndex maxchars;
		Boolean cfOK;
	
		cfstr = CFStringCreateWithCString(kCFAllocatorDefault, defaults->volumeName, kCFStringEncodingUTF8);

		/* Find out what Mac encoding to use: */
		maxchars = MIN(sizeof(unibuf)/sizeof(UniChar), CFStringGetLength(cfstr));
		CFStringGetCharacters(cfstr, CFRangeMake(0, maxchars), unibuf);
		cfOK = CFStringGetPascalString(cfstr, mdbp->drVN, sizeof(mdbp->drVN), GetDefaultEncoding());
		CFRelease(cfstr);

		if (!cfOK) {
			mdbp->drVN[0] = strlen(kDefaultVolumeNameStr);
			bcopy(kDefaultVolumeNameStr, &mdbp->drVN[1], mdbp->drVN[0]);
			warnx("invalid HFS name: \"%s\", using \"%s\" instead",
			      defaults->volumeName, kDefaultVolumeNameStr);
		}
		/* defaults->volumeName is used later for the root dir key */
		bcopy(&mdbp->drVN[1], defaults->volumeName, mdbp->drVN[0]);
		defaults->volumeName[mdbp->drVN[0]] = '\0';
	}
#else
	mdbp->drVN[0] = strlen(defaults->volumeName);
	bcopy(defaults->volumeName, &mdbp->drVN[1], mdbp->drVN[0]);
#endif
	
	SET_SWAP_BE32(mdbp->drWrCnt, kWriteSeqNum);

	SET_SWAP_BE32(mdbp->drXTFlSize, defaults->extentsClumpSize);
	SET_SWAP_BE32(mdbp->drXTClpSiz, defaults->extentsClumpSize);
	SET_SWAP_BE16(mdbp->drXTExtRec[0].startBlock, 0);
	SET_SWAP_BE16(mdbp->drXTExtRec[0].blockCount, SWAP_BE32(mdbp->drXTFlSize) / alBlkSize);
	SET_SWAP_BE16(mdbp->drFreeBks, SWAP_BE16(mdbp->drFreeBks) - SWAP_BE16(mdbp->drXTExtRec[0].blockCount));

	SET_SWAP_BE32(mdbp->drCTFlSize, defaults->catalogClumpSize);
	SET_SWAP_BE32(mdbp->drCTClpSiz, defaults->catalogClumpSize);
	SET_SWAP_BE16(mdbp->drCTExtRec[0].startBlock, SWAP_BE16(mdbp->drXTExtRec[0].startBlock) +
					 SWAP_BE16(mdbp->drXTExtRec[0].blockCount));
	SET_SWAP_BE16(mdbp->drCTExtRec[0].blockCount, SWAP_BE32(mdbp->drCTFlSize) / alBlkSize);
	SET_SWAP_BE16(mdbp->drFreeBks, SWAP_BE16(mdbp->drFreeBks) - SWAP_BE16(mdbp->drCTExtRec[0].blockCount));

	if (defaults->flags & kMakeHFSWrapper) {
		SET_SWAP_BE32(mdbp->drFilCnt, kWapperFileCount);
		SET_SWAP_BE16(mdbp->drNmFls, kWapperFileCount);
		SET_SWAP_BE32(mdbp->drNxtCNID, SWAP_BE32(mdbp->drNxtCNID) + kWapperFileCount);

		/* set blessed system folder to be root folder (2) */
		SET_SWAP_BE32(mdbp->drFndrInfo[0] ,kHFSRootFolderID);

		SET_SWAP_BE16(mdbp->drEmbedSigWord, kHFSPlusSigWord);

		/* software lock it and tag as having "bad" blocks */
		SET_SWAP_BE16(mdbp->drAtrb, SWAP_BE16(mdbp->drAtrb) | kHFSVolumeSparedBlocksMask | kHFSVolumeSoftwareLockMask);
	}
}


/*
 * WriteVH
 *
 * Writes the Volume Header (VH) to disk.
 *
 * The VH is byte-swapped if necessary to big endian. Since this
 * is always the last operation, there's no point in unswapping it.
 */
static void
WriteVH (const DriveInfo *driveInfo, HFSPlusVolumeHeader *hp)
{
	//SWAP_HFSPLUSVH (hp);

	WriteBuffer(driveInfo, 2, kBytesPerSector, hp);
	WriteBuffer(driveInfo, driveInfo->totalSectors - 2, kBytesPerSector, hp);
}


/*
 * InitVH
 *
 * Initialize a Volume Header record.
 */
static void
InitVH(hfsparams_t *defaults, UInt64 sectors, HFSPlusVolumeHeader *hp)
{
	UInt32	blockSize;
	UInt32	blockCount;
	UInt32	blocksUsed;
	UInt32	bitmapBlocks;
	UInt16	burnedBlocksBeforeVH = 0;
	UInt16	burnedBlocksAfterAltVH = 0;
	UInt32  nextBlock;
	
	bzero(hp, kBytesPerSector);

	blockSize = defaults->blockSize;
	blockCount = sectors / (blockSize >> kLog2SectorSize);

	/*
	 * HFSPlusVolumeHeader is located at sector 2, so we may need
	 * to invalidate blocks before HFSPlusVolumeHeader.
	 */
	if ( blockSize == 512 ) {
		burnedBlocksBeforeVH = 2;		/* 2 before VH */
		burnedBlocksAfterAltVH = 1;		/* 1 after altVH */
	} else if ( blockSize == 1024 ) {
		burnedBlocksBeforeVH = 1;
	}

	bitmapBlocks = defaults->allocationClumpSize / blockSize;

	/* note: add 2 for the Alternate VH, and VH */
	blocksUsed = 2 + burnedBlocksBeforeVH + burnedBlocksAfterAltVH + bitmapBlocks;

	SET_SWAP_BE16(hp->signature, kHFSPlusSigWord);
	SET_SWAP_BE16(hp->version, kHFSPlusVersion);
	SET_SWAP_BE32(hp->attributes, kHFSVolumeUnmountedMask);
	SET_SWAP_BE32(hp->lastMountedVersion, kHFSPlusMountVersion);

	/* NOTE: create date is in local time, not GMT!  */
	SET_SWAP_BE32(hp->createDate, UTCToLocal(defaults->createDate));
	SET_SWAP_BE32(hp->modifyDate, defaults->createDate);
	SET_SWAP_BE32(hp->backupDate, 0);
	SET_SWAP_BE32(hp->checkedDate, defaults->createDate);

//	hp->fileCount = 0;
//	hp->folderCount = 0;

	SET_SWAP_BE32(hp->blockSize, blockSize);
	SET_SWAP_BE32(hp->totalBlocks, blockCount);
	SET_SWAP_BE32(hp->freeBlocks, blockCount);	/* will be adjusted at the end */

	SET_SWAP_BE32(hp->rsrcClumpSize, defaults->rsrcClumpSize);
	SET_SWAP_BE32(hp->dataClumpSize, defaults->dataClumpSize);
	SET_SWAP_BE32(hp->nextCatalogID, defaults->nextFreeFileID);
	SET_SWAP_BE64(hp->encodingsBitmap, 1);	/* just set to MacRoman */

	/* set up allocation bitmap file */
	SET_SWAP_BE32(hp->allocationFile.clumpSize, defaults->allocationClumpSize);
	SET_SWAP_BE64(hp->allocationFile.logicalSize, defaults->allocationClumpSize);
	SET_SWAP_BE32(hp->allocationFile.totalBlocks, bitmapBlocks);
  	SET_SWAP_BE32(hp->allocationFile.extents[0].startBlock, 1 + burnedBlocksBeforeVH);
	SET_SWAP_BE32(hp->allocationFile.extents[0].blockCount, bitmapBlocks);
	
	/* set up journal files */
	if (defaults->journaledHFS) {
		SET_SWAP_BE32(hp->fileCount, 2);
		SET_SWAP_BE32(hp->attributes, SWAP_BE32(hp->attributes) | kHFSVolumeJournaledMask);
		SET_SWAP_BE32(hp->lastMountedVersion, kHFSJMountVersion);
		SET_SWAP_BE32(hp->nextCatalogID, SWAP_BE32(hp->nextCatalogID) + 2);

		/*
		 * Allocate 1 block for the journalInfoBlock the
		 * journal file size is pass in hfsparams_t
		 */
	    SET_SWAP_BE32(hp->journalInfoBlock, SWAP_BE32(hp->allocationFile.extents[0].startBlock) +
				   SWAP_BE32(hp->allocationFile.extents[0].blockCount));
	    blocksUsed += 1 + ((defaults->journalSize) / blockSize);
	    nextBlock = SWAP_BE32(hp->journalInfoBlock) + 1 + (defaults->journalSize / blockSize);
	} else {
	    SET_SWAP_BE32(hp->journalInfoBlock, 0);
	    nextBlock = SWAP_BE32(hp->allocationFile.extents[0].startBlock) +
			SWAP_BE32(hp->allocationFile.extents[0].blockCount);
	}

	/* set up extents b-tree file */
	SET_SWAP_BE32(hp->extentsFile.clumpSize, defaults->extentsClumpSize);
	SET_SWAP_BE64(hp->extentsFile.logicalSize, defaults->extentsClumpSize);
	SET_SWAP_BE32(hp->extentsFile.totalBlocks, defaults->extentsClumpSize / blockSize);
	SET_SWAP_BE32(hp->extentsFile.extents[0].startBlock, nextBlock);
	SET_SWAP_BE32(hp->extentsFile.extents[0].blockCount, SWAP_BE32(hp->extentsFile.totalBlocks));
	blocksUsed += SWAP_BE32(hp->extentsFile.totalBlocks);


	/* set up catalog b-tree file */
	SET_SWAP_BE32(hp->catalogFile.clumpSize, defaults->catalogClumpSize);
	SET_SWAP_BE64(hp->catalogFile.logicalSize, defaults->catalogClumpSize);
	SET_SWAP_BE32(hp->catalogFile.totalBlocks, defaults->catalogClumpSize / blockSize);
	SET_SWAP_BE32(hp->catalogFile.extents[0].startBlock,
		SWAP_BE32(hp->extentsFile.extents[0].startBlock) +
		SWAP_BE32(hp->extentsFile.extents[0].blockCount));
	SET_SWAP_BE32(hp->catalogFile.extents[0].blockCount, SWAP_BE32(hp->catalogFile.totalBlocks));
	blocksUsed += SWAP_BE32(hp->catalogFile.totalBlocks);

	SET_SWAP_BE32(hp->freeBlocks, SWAP_BE32(hp->freeBlocks) - blocksUsed);
	
	/*
	 * Add some room for the catalog file to grow...
	 */
	SET_SWAP_BE32(hp->nextAllocation, blocksUsed - 1 - burnedBlocksAfterAltVH +
		10 * (SWAP_BE32(hp->catalogFile.clumpSize) / SWAP_BE32(hp->blockSize)));
}


/*
 * InitBitmap
 * 	
 * This routine initializes the Allocation Bitmap. Allocation blocks
 * that are in use have their corresponding bit set.
 * 
 * It assumes that initially there are no gaps between allocated blocks.
 * 
 * It also assumes the buffer is big enough to hold all the bits
 * (ie its at least (alBlksUsed/8) bytes in size.
 */
static void
WriteBitmap(const DriveInfo *driveInfo, UInt32 startingSector,
        UInt32 alBlksUsed, UInt8 *buffer)
{
	UInt32	bytes, bits, bytesUsed;

	bytes = alBlksUsed >> 3;
	bits  = alBlksUsed & 0x0007;

	(void)memset(buffer, 0xFF, bytes);

	if (bits) {
		*(UInt8 *)(buffer + bytes) = (0xFF00 >> bits) & 0xFF;
		++bytes;
	}

	bytesUsed = ROUNDUP(bytes, driveInfo->sectorSize);

	if (bytesUsed > bytes)
		bzero(buffer + bytes, bytesUsed - bytes);
	WriteBuffer(driveInfo, startingSector, bytesUsed, buffer);
}


/*
 * WriteExtentsFile
 *
 * Initializes and writes out the extents b-tree file.
 *
 * Byte swapping is performed in place. The buffer should not be
 * accessed through direct casting once it leaves this function.
 */
static void
WriteExtentsFile(const DriveInfo *driveInfo, UInt32 startingSector,
        const hfsparams_t *dp, HFSExtentDescriptor *bbextp, void *buffer,
        UInt32 *bytesUsed, UInt32 *mapNodes)
{
	BTNodeDescriptor	*ndp;
	BTHeaderRec		*bthp;
	UInt8			*bmp;
	UInt32			nodeBitsInHeader;
	UInt32			fileSize;
	UInt32			nodeSize;
	UInt32			temp;
	SInt16			offset;
	int			wrapper = (dp->flags & kMakeHFSWrapper);

	*mapNodes = 0;
	fileSize = dp->extentsClumpSize;
	nodeSize = dp->extentsNodeSize;

	bzero(buffer, nodeSize);


	/* FILL IN THE NODE DESCRIPTOR:  */
	ndp = (BTNodeDescriptor *)buffer;
	ndp->kind			= kBTHeaderNode;
	SET_SWAP_BE16(ndp->numRecords, 3);
	offset = sizeof(BTNodeDescriptor);

	SETOFFSET(buffer, nodeSize, offset, 1);


	/* FILL IN THE HEADER RECORD:  */
	bthp = (BTHeaderRec *)((UInt8 *)buffer + offset);
    //	bthp->treeDepth		= 0;
    //	bthp->rootNode		= 0;
    //	bthp->firstLeafNode	= 0;
    //	bthp->lastLeafNode	= 0;
    //	bthp->leafRecords	= 0;
	SET_SWAP_BE16(bthp->nodeSize, nodeSize);
	SET_SWAP_BE32(bthp->totalNodes, fileSize / nodeSize);
	SET_SWAP_BE32(bthp->freeNodes, SWAP_BE32(bthp->totalNodes) - 1);  /* header */
	SET_SWAP_BE32(bthp->clumpSize, fileSize);

	if (dp->signature == kHFSPlusSigWord) {
		SET_SWAP_BE32(bthp->attributes, SWAP_BE32(bthp->attributes) | kBTBigKeysMask);
		SET_SWAP_BE16(bthp->maxKeyLength, kHFSPlusExtentKeyMaximumLength);
	} else {
		SET_SWAP_BE16(bthp->maxKeyLength, kHFSExtentKeyMaximumLength);

		/* wrapper has a bad-block extent record */
		if (wrapper) {
			SET_SWAP_BE16(bthp->treeDepth, SWAP_BE16(bthp->treeDepth) + 1);
			SET_SWAP_BE32(bthp->leafRecords, SWAP_BE32(bthp->leafRecords) + 1);
			SET_SWAP_BE32(bthp->rootNode, 1);
			SET_SWAP_BE32(bthp->firstLeafNode, 1);
			SET_SWAP_BE32(bthp->lastLeafNode, 1);
			SET_SWAP_BE32(bthp->freeNodes, SWAP_BE32(bthp->freeNodes) - 1);
		}
	}
	offset += sizeof(BTHeaderRec);

	SETOFFSET(buffer, nodeSize, offset, 2);

	offset += kBTreeHeaderUserBytes;

	SETOFFSET(buffer, nodeSize, offset, 3);


	/* FIGURE OUT HOW MANY MAP NODES (IF ANY):  */
	nodeBitsInHeader = 8 * (nodeSize
					- sizeof(BTNodeDescriptor)
					- sizeof(BTHeaderRec)
					- kBTreeHeaderUserBytes
					- (4 * sizeof(SInt16)) );

	if (SWAP_BE32 (bthp->totalNodes) > nodeBitsInHeader) {
		UInt32	nodeBitsInMapNode;
		
		SET_SWAP_BE32(ndp->fLink, SWAP_BE32 (bthp->lastLeafNode) + 1);
		nodeBitsInMapNode = 8 * (nodeSize
						- sizeof(BTNodeDescriptor)
						- (2 * sizeof(SInt16))
						- 2 );
		*mapNodes = (SWAP_BE32 (bthp->totalNodes) - nodeBitsInHeader +
			(nodeBitsInMapNode - 1)) / nodeBitsInMapNode;
		SET_SWAP_BE32(bthp->freeNodes, SWAP_BE32 (bthp->freeNodes) - *mapNodes);
	}


	/* 
	 * FILL IN THE MAP RECORD, MARKING NODES THAT ARE IN USE.
	 * Note - worst case (32MB alloc blk) will have only 18 nodes in use.
	 */
	bmp = ((UInt8 *)buffer + offset);
	temp = SWAP_BE32 (bthp->totalNodes) - SWAP_BE32 (bthp->freeNodes);

	/* Working a byte at a time is endian safe */
	while (temp >= 8) { *bmp = 0xFF; temp -= 8; bmp++; }
	*bmp = ~(0xFF >> temp);
	offset += nodeBitsInHeader/8;

	SETOFFSET(buffer, nodeSize, offset, 4);

	if (wrapper) {
		InitExtentsRoot(nodeSize, bbextp, (buffer + nodeSize));
	}
	
	*bytesUsed = (SWAP_BE32 (bthp->totalNodes) - SWAP_BE32 (bthp->freeNodes) - *mapNodes) * nodeSize;

	WriteBuffer(driveInfo, startingSector, *bytesUsed, buffer);
}


static void
InitExtentsRoot(UInt16 btNodeSize, HFSExtentDescriptor *bbextp, void *buffer)
{
	BTNodeDescriptor	*ndp;
	HFSExtentKey		*ekp;
	HFSExtentRecord		*edp;
	SInt16				offset;

	bzero(buffer, btNodeSize);

	/*
	 * All nodes have a node descriptor...
	 */
	ndp = (BTNodeDescriptor *)buffer;
	ndp->kind			= kBTLeafNode;
	SET_SWAP_BE16(ndp->numRecords, 1);
	ndp->height			= 1;
	offset = sizeof(BTNodeDescriptor);

	SETOFFSET(buffer, btNodeSize, offset, 1);

	/*
	 * First and only record is bad block extents...
	 */
	ekp = (HFSExtentKey *)((UInt8 *) buffer + offset);
	ekp->keyLength		= kHFSExtentKeyMaximumLength;
	// ekp->forkType	= 0;
	SET_SWAP_BE32(ekp->fileID, kHFSBadBlockFileID);
	// ekp->startBlock	= 0;
	offset += sizeof(HFSExtentKey);

	edp = (HFSExtentRecord *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(edp[0]->startBlock, SWAP_BE16(bbextp->startBlock));
	SET_SWAP_BE16(edp[0]->blockCount, SWAP_BE16(bbextp->blockCount));
	offset += sizeof(HFSExtentRecord);

	SETOFFSET(buffer, btNodeSize, offset, 2);
}

static void
WriteJournalInfo(const DriveInfo *driveInfo, UInt32 startingSector,
		 const hfsparams_t *dp, HFSPlusVolumeHeader *header,
		 void *buffer)
{
    JournalInfoBlock *jibp = buffer;
    
    memset(buffer, 0xdb, SWAP_BE32(header->blockSize));
    memset(jibp, 0, sizeof(JournalInfoBlock));
    
    SET_SWAP_BE32(jibp->flags, kJIJournalInFSMask | kJIJournalNeedInitMask);
    SET_SWAP_BE64(jibp->offset, (SWAP_BE32(header->journalInfoBlock) + 1) * SWAP_BE32(header->blockSize));
    SET_SWAP_BE64(jibp->size, dp->journalSize);

    WriteBuffer(driveInfo, startingSector, SWAP_BE32(header->blockSize), buffer);
}


/*
 * WriteCatalogFile
 *	
 * This routine initializes a Catalog B-Tree.
 *
 * Note: Since large volumes can have bigger b-trees they
 * might need to have map nodes setup.
 */
static void
WriteCatalogFile(const DriveInfo *driveInfo, UInt32 startingSector,
        const hfsparams_t *dp, HFSPlusVolumeHeader *header, void *buffer,
        UInt32 *bytesUsed, UInt32 *mapNodes)
{
	BTNodeDescriptor	*ndp;
	BTHeaderRec		*bthp;
	UInt8			*bmp;
	UInt32			nodeBitsInHeader;
	UInt32			fileSize;
	UInt32			nodeSize;
	UInt32			temp;
	SInt16			offset;
	int			wrapper = (dp->flags & kMakeHFSWrapper);

	*mapNodes = 0;
	fileSize = dp->catalogClumpSize;
	nodeSize = dp->catalogNodeSize;

	bzero(buffer, nodeSize);


	/* FILL IN THE NODE DESCRIPTOR:  */
	ndp = (BTNodeDescriptor *)buffer;
	ndp->kind			= kBTHeaderNode;
	SET_SWAP_BE16(ndp->numRecords, 3);
	offset = sizeof(BTNodeDescriptor);

	SETOFFSET(buffer, nodeSize, offset, 1);


	/* FILL IN THE HEADER RECORD:  */
	bthp = (BTHeaderRec *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(bthp->treeDepth, 1);
	SET_SWAP_BE32(bthp->rootNode, 1);
	SET_SWAP_BE32(bthp->firstLeafNode, 1);
	SET_SWAP_BE32(bthp->lastLeafNode, 1);
	SET_SWAP_BE32(bthp->leafRecords, dp->journaledHFS ? 6 : 2);
	SET_SWAP_BE16(bthp->nodeSize, nodeSize);
	SET_SWAP_BE32(bthp->totalNodes, fileSize / nodeSize);
	SET_SWAP_BE32(bthp->freeNodes, SWAP_BE32 (bthp->totalNodes) - 2);  /* header and root */
	SET_SWAP_BE32(bthp->clumpSize, fileSize);

	if (dp->signature == kHFSPlusSigWord) {
		SET_SWAP_BE32(bthp->attributes, SWAP_BE32(bthp->attributes) | kBTVariableIndexKeysMask | kBTBigKeysMask);
		SET_SWAP_BE16(bthp->maxKeyLength, kHFSPlusCatalogKeyMaximumLength);
	} else {
		SET_SWAP_BE16(bthp->maxKeyLength, kHFSCatalogKeyMaximumLength);

		if (dp->flags & kMakeHFSWrapper) {
			SET_SWAP_BE16(bthp->treeDepth, SWAP_BE16 (bthp->treeDepth) + 1);
			SET_SWAP_BE32(bthp->leafRecords, SWAP_BE32 (bthp->leafRecords) + kWapperFileCount);
			SET_SWAP_BE32(bthp->firstLeafNode, SWAP_BE32 (bthp->rootNode) + 1);
			SET_SWAP_BE32(bthp->lastLeafNode, SWAP_BE32 (bthp->firstLeafNode) + 1);
			SET_SWAP_BE32(bthp->freeNodes, SWAP_BE32 (bthp->freeNodes) - 2);  /* tree now split with 2 leaf nodes */
		}
	}
	offset += sizeof(BTHeaderRec);

	SETOFFSET(buffer, nodeSize, offset, 2);

	offset += kBTreeHeaderUserBytes;

	SETOFFSET(buffer, nodeSize, offset, 3);

	/* FIGURE OUT HOW MANY MAP NODES (IF ANY):  */
	nodeBitsInHeader = 8 * (nodeSize
					- sizeof(BTNodeDescriptor)
					- sizeof(BTHeaderRec)
					- kBTreeHeaderUserBytes
					- (4 * sizeof(SInt16)) );

	if (SWAP_BE32 (bthp->totalNodes) > nodeBitsInHeader) {
		UInt32	nodeBitsInMapNode;
		
		SET_SWAP_BE32(ndp->fLink, SWAP_BE32 (bthp->lastLeafNode) + 1);
		nodeBitsInMapNode = 8 * (nodeSize
						- sizeof(BTNodeDescriptor)
						- (2 * sizeof(SInt16))
						- 2 );
		*mapNodes = (SWAP_BE32 (bthp->totalNodes) - nodeBitsInHeader +
			(nodeBitsInMapNode - 1)) / nodeBitsInMapNode;
		SET_SWAP_BE32(bthp->freeNodes, SWAP_BE32 (bthp->freeNodes) - *mapNodes);
	}

	/* 
	 * FILL IN THE MAP RECORD, MARKING NODES THAT ARE IN USE.
	 * Note - worst case (32MB alloc blk) will have only 18 nodes in use.
	 */
	bmp = ((UInt8 *)buffer + offset);
	temp = SWAP_BE32 (bthp->totalNodes) - SWAP_BE32 (bthp->freeNodes);

	/* Working a byte at a time is endian safe */
	while (temp >= 8) { *bmp = 0xFF; temp -= 8; bmp++; }
	*bmp = ~(0xFF >> temp);
	offset += nodeBitsInHeader/8;

	SETOFFSET(buffer, nodeSize, offset, 4);

	if (dp->signature == kHFSPlusSigWord) {
		InitCatalogRoot_HFSPlus(dp, header, buffer + nodeSize);

	} else if (wrapper) {
		InitCatalogRoot_HFS  (dp, buffer + (1 * nodeSize));
		InitFirstCatalogLeaf (dp, buffer + (2 * nodeSize), TRUE);
		InitSecondCatalogLeaf(dp, buffer + (3 * nodeSize));

	} else /* plain HFS */ {
		InitFirstCatalogLeaf(dp, buffer + nodeSize, FALSE);
	}

	*bytesUsed = (SWAP_BE32 (bthp->totalNodes) - SWAP_BE32 (bthp->freeNodes) - *mapNodes) * nodeSize;

	WriteBuffer(driveInfo, startingSector, *bytesUsed, buffer);
}


static void
InitCatalogRoot_HFSPlus(const hfsparams_t *dp, const HFSPlusVolumeHeader *header, void * buffer)
{
	BTNodeDescriptor		*ndp;
	HFSPlusCatalogKey		*ckp;
	HFSPlusCatalogKey		*tkp;
	HFSPlusCatalogFolder	*cdp;
	HFSPlusCatalogFile		*cfp;
	HFSPlusCatalogThread	*ctp;
	UInt16					nodeSize;
	SInt16					offset;
	UInt32					unicodeBytes;
#if __APPLE__
	UInt8 canonicalName[256];
	CFStringRef cfstr;
	Boolean	cfOK;
#endif
	int index = 0;

	nodeSize = dp->catalogNodeSize;
	bzero(buffer, nodeSize);

	/*
	 * All nodes have a node descriptor...
	 */
	ndp = (BTNodeDescriptor *)buffer;
	ndp->kind   = kBTLeafNode;
	ndp->height = 1;
	SET_SWAP_BE16(ndp->numRecords, dp->journaledHFS ? 6 : 2);
	offset = sizeof(BTNodeDescriptor);

	SETOFFSET(buffer, nodeSize, offset, ++index);

	/*
	 * First record is always the root directory...
	 */
	ckp = (HFSPlusCatalogKey *)((UInt8 *)buffer + offset);

#if __APPLE__
	/* Use CFString functions to get a HFSPlus Canonical name */
	cfstr = CFStringCreateWithCString(kCFAllocatorDefault, dp->volumeName, kCFStringEncodingUTF8);
	cfOK = _CFStringGetFileSystemRepresentation(cfstr, canonicalName, sizeof(canonicalName));

	if (!cfOK || ConvertUTF8toUnicode(canonicalName, sizeof(ckp->nodeName.unicode),
		ckp->nodeName.unicode, &ckp->nodeName.length)) {

		/* On conversion errors "untitled" is used as a fallback. */
		(void) ConvertUTF8toUnicode(kDefaultVolumeNameStr,
									sizeof(ckp->nodeName.unicode),
									ckp->nodeName.unicode,
									&ckp->nodeName.length);
		warnx("invalid HFS+ name: \"%s\", using \"%s\" instead",
		      dp->volumeName, kDefaultVolumeNameStr);
	}
	CFRelease(cfstr);
#else
	ConvertUTF8toUnicode(dp->volumeName, sizeof(ckp->nodeName.unicode),
					ckp->nodeName.unicode, &ckp->nodeName.length);
#endif

	unicodeBytes = sizeof(UniChar) * SWAP_BE16 (ckp->nodeName.length);

	SET_SWAP_BE16(ckp->keyLength, kHFSPlusCatalogKeyMinimumLength + unicodeBytes);
	SET_SWAP_BE32(ckp->parentID, kHFSRootParentID);
	offset += SWAP_BE16 (ckp->keyLength) + 2;

	cdp = (HFSPlusCatalogFolder *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(cdp->recordType, kHFSPlusFolderRecord);
	SET_SWAP_BE32(cdp->valence, dp->journaledHFS ? 2 : 0);
	SET_SWAP_BE32(cdp->folderID, kHFSRootFolderID);
	SET_SWAP_BE32(cdp->createDate, dp->createDate);
	SET_SWAP_BE32(cdp->contentModDate, dp->createDate);
	SET_SWAP_BE32(cdp->textEncoding, GetDefaultEncoding());
	offset += sizeof(HFSPlusCatalogFolder);

	SETOFFSET(buffer, nodeSize, offset, ++index);

	/*
	 * Second record is always the root directory thread...
	 */
	tkp = (HFSPlusCatalogKey *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(tkp->keyLength, kHFSPlusCatalogKeyMinimumLength);
	SET_SWAP_BE32(tkp->parentID, kHFSRootFolderID);
	// tkp->nodeName.length = 0;

	offset += SWAP_BE16 (tkp->keyLength) + 2;

	ctp = (HFSPlusCatalogThread *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(ctp->recordType, kHFSPlusFolderThreadRecord);
	SET_SWAP_BE32(ctp->parentID, kHFSRootParentID);
	bcopy(&ckp->nodeName, &ctp->nodeName, sizeof(UInt16) + unicodeBytes);
	offset += (sizeof(HFSPlusCatalogThread)
			- (sizeof(ctp->nodeName.unicode) - unicodeBytes) );

	SETOFFSET(buffer, nodeSize, offset, ++index);
	
	/*
	 * Add records for ".journal" and ".journal_info_block" files:
	 */
	if (dp->journaledHFS) {
		struct HFSUniStr255 *nodename1, *nodename2;

		/* File record #1 */
		ckp = (HFSPlusCatalogKey *)((UInt8 *)buffer + offset);
		(void) ConvertUTF8toUnicode(HFS_JOURNAL_FILE, sizeof(ckp->nodeName.unicode),
		                            ckp->nodeName.unicode, &ckp->nodeName.length);
		unicodeBytes = sizeof(UniChar) * SWAP_BE16 (ckp->nodeName.length);
		SET_SWAP_BE16(ckp->keyLength, kHFSPlusCatalogKeyMinimumLength + unicodeBytes);
		SET_SWAP_BE32(ckp->parentID, kHFSRootFolderID);
		offset += SWAP_BE16 (ckp->keyLength) + 2;
	
		cfp = (HFSPlusCatalogFile *)((UInt8 *)buffer + offset);
		SET_SWAP_BE16(cfp->recordType, kHFSPlusFileRecord);
		SET_SWAP_BE32(cfp->fileID, dp->nextFreeFileID);
		SET_SWAP_BE32(cfp->createDate, dp->createDate + 1);
		SET_SWAP_BE32(cfp->contentModDate, dp->createDate + 1);
		SET_SWAP_BE32(cfp->textEncoding, 0);

		SET_SWAP_BE16(cfp->bsdInfo.fileMode, S_IFREG);
		cfp->bsdInfo.ownerFlags = UF_NODUMP;
		SET_SWAP_BE32(cfp->userInfo.fdType, kJournalFileType);
		SET_SWAP_BE32(cfp->userInfo.fdCreator, kHFSPlusCreator);
		SET_SWAP_BE16(cfp->userInfo.fdFlags, kIsInvisible + kNameLocked);
		SET_SWAP_BE64(cfp->dataFork.logicalSize, dp->journalSize);
		SET_SWAP_BE32(cfp->dataFork.totalBlocks, dp->journalSize / dp->blockSize);

		SET_SWAP_BE32(cfp->dataFork.extents[0].startBlock, SWAP_BE32(header->journalInfoBlock) + 1);
		cfp->dataFork.extents[0].blockCount = cfp->dataFork.totalBlocks;

		offset += sizeof(HFSPlusCatalogFile);
		SETOFFSET(buffer, nodeSize, offset, ++index);
		nodename1 = &ckp->nodeName;

		/* File record #2 */
		ckp = (HFSPlusCatalogKey *)((UInt8 *)buffer + offset);
		(void) ConvertUTF8toUnicode(HFS_JOURNAL_INFO, sizeof(ckp->nodeName.unicode),
		                            ckp->nodeName.unicode, &ckp->nodeName.length);
		unicodeBytes = sizeof(UniChar) * SWAP_BE16 (ckp->nodeName.length);
		SET_SWAP_BE16(ckp->keyLength, kHFSPlusCatalogKeyMinimumLength + unicodeBytes);
		SET_SWAP_BE32(ckp->parentID, kHFSRootFolderID);
		offset += SWAP_BE16 (ckp->keyLength) + 2;
	
		cfp = (HFSPlusCatalogFile *)((UInt8 *)buffer + offset);
		SET_SWAP_BE16(cfp->recordType, kHFSPlusFileRecord);
		SET_SWAP_BE32(cfp->fileID, dp->nextFreeFileID + 1);
		SET_SWAP_BE32(cfp->createDate, dp->createDate);
		SET_SWAP_BE32(cfp->contentModDate, dp->createDate);
		SET_SWAP_BE32(cfp->textEncoding, 0);

		SET_SWAP_BE16(cfp->bsdInfo.fileMode, S_IFREG);
		cfp->bsdInfo.ownerFlags = UF_NODUMP;
		SET_SWAP_BE32(cfp->userInfo.fdType, kJournalFileType);
		SET_SWAP_BE32(cfp->userInfo.fdCreator, kHFSPlusCreator);
		SET_SWAP_BE16(cfp->userInfo.fdFlags, kIsInvisible + kNameLocked);
		SET_SWAP_BE64(cfp->dataFork.logicalSize, dp->blockSize);
		SET_SWAP_BE32(cfp->dataFork.totalBlocks, 1);

		cfp->dataFork.extents[0].startBlock = header->journalInfoBlock;
		cfp->dataFork.extents[0].blockCount = cfp->dataFork.totalBlocks;

		offset += sizeof(HFSPlusCatalogFile);
		SETOFFSET(buffer, nodeSize, offset, ++index);
		nodename2 = &ckp->nodeName;

		/* Thread record for file #1 */
		tkp = (HFSPlusCatalogKey *)((UInt8 *)buffer + offset);
		SET_SWAP_BE16(tkp->keyLength, kHFSPlusCatalogKeyMinimumLength);
		SET_SWAP_BE32(tkp->parentID, dp->nextFreeFileID);
		SET_SWAP_BE16(tkp->nodeName.length, 0);
		offset += SWAP_BE16 (tkp->keyLength) + 2;
	
		ctp = (HFSPlusCatalogThread *)((UInt8 *)buffer + offset);
		SET_SWAP_BE16(ctp->recordType, kHFSPlusFileThreadRecord);
		SET_SWAP_BE32(ctp->parentID, kHFSRootFolderID);
		bcopy(nodename1, &ctp->nodeName, sizeof(UInt16) + unicodeBytes);
		offset += (sizeof(HFSPlusCatalogThread)
				- (sizeof(ctp->nodeName.unicode) - unicodeBytes) );
		SETOFFSET(buffer, nodeSize, offset, ++index);

		/* Thread record for file #2 */
		tkp = (HFSPlusCatalogKey *)((UInt8 *)buffer + offset);
		SET_SWAP_BE16(tkp->keyLength, kHFSPlusCatalogKeyMinimumLength);
		SET_SWAP_BE32(tkp->parentID, dp->nextFreeFileID + 1);
		SET_SWAP_BE16(tkp->nodeName.length, 0);
		offset += SWAP_BE16 (tkp->keyLength) + 2;
	
		ctp = (HFSPlusCatalogThread *)((UInt8 *)buffer + offset);
		SET_SWAP_BE16(ctp->recordType, kHFSPlusFileThreadRecord);
		SET_SWAP_BE32(ctp->parentID, kHFSRootFolderID);
		bcopy(nodename2, &ctp->nodeName, sizeof(UInt16) + unicodeBytes);
		offset += (sizeof(HFSPlusCatalogThread)
				- (sizeof(ctp->nodeName.unicode) - unicodeBytes) );
		SETOFFSET(buffer, nodeSize, offset, ++index);
	}
}


static void
InitFirstCatalogLeaf(const hfsparams_t *dp, void * buffer, int wrapper)
{
	BTNodeDescriptor	*ndp;
	HFSCatalogKey		*ckp;
	HFSCatalogKey		*tkp;
	HFSCatalogFolder	*cdp;
	HFSCatalogFile		*cfp;
	HFSCatalogThread	*ctp;
	UInt16			nodeSize;
	SInt16			offset;
	UInt32			timeStamp;

	nodeSize = dp->catalogNodeSize;
	timeStamp = UTCToLocal(dp->createDate);
	bzero(buffer, nodeSize);

	/*
	 * All nodes have a node descriptor...
	 */
	ndp = (BTNodeDescriptor *)buffer;
	ndp->kind			= kBTLeafNode;
	SET_SWAP_BE16(ndp->numRecords, 2);
	ndp->height			= 1;
	offset = sizeof(BTNodeDescriptor);

	SETOFFSET(buffer, nodeSize, offset, 1);

	/*
	 * First record is always the root directory...
	 */
	ckp = (HFSCatalogKey *)((UInt8 *)buffer + offset);
	ckp->nodeName[0]	= strlen(dp->volumeName);
	bcopy(dp->volumeName, &ckp->nodeName[1], ckp->nodeName[0]);
	ckp->keyLength		= 1 + 4 + ((ckp->nodeName[0] + 2) & 0xFE);  /* pad to word */
	SET_SWAP_BE32(ckp->parentID, kHFSRootParentID);
	offset += ckp->keyLength + 1;

	cdp = (HFSCatalogFolder *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(cdp->recordType, kHFSFolderRecord);
	if (wrapper)
		SET_SWAP_BE16(cdp->valence, SWAP_BE16 (cdp->valence) + kWapperFileCount);
	SET_SWAP_BE32(cdp->folderID, kHFSRootFolderID);
	SET_SWAP_BE32(cdp->createDate, timeStamp);
	SET_SWAP_BE32(cdp->modifyDate, timeStamp);
	offset += sizeof(HFSCatalogFolder);

	SETOFFSET(buffer, nodeSize, offset, 2);

	/*
	 * Second record is always the root directory thread...
	 */
	tkp = (HFSCatalogKey *)((UInt8 *)buffer + offset);
	tkp->keyLength		= kHFSCatalogKeyMinimumLength;
	SET_SWAP_BE32(tkp->parentID, kHFSRootFolderID);
	// tkp->nodeName[0] = 0;
	offset += tkp->keyLength + 2;

	ctp = (HFSCatalogThread *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(ctp->recordType, kHFSFolderThreadRecord);
	SET_SWAP_BE32(ctp->parentID, kHFSRootParentID);
	bcopy(ckp->nodeName, ctp->nodeName, ckp->nodeName[0]+1);
	offset += sizeof(HFSCatalogThread);

	SETOFFSET(buffer, nodeSize, offset, 3);
	
	/*
	 * For Wrapper volumes there are more file records...
	 */
	if (wrapper) {
		SET_SWAP_BE32(ndp->fLink, 3);
		SET_SWAP_BE16(ndp->numRecords, SWAP_BE16 (ndp->numRecords) + 2);

		/*
		 * Add "Desktop DB" file...
		 */
		ckp = (HFSCatalogKey *)((UInt8 *)buffer + offset);
		ckp->keyLength			= 1 + 4 + ((kDTDB_Chars + 2) & 0xFE);  /* pad to word */
		SET_SWAP_BE32(ckp->parentID, kHFSRootFolderID);
		ckp->nodeName[0]		= kDTDB_Chars;
		bcopy(kDTDB_Name, &ckp->nodeName[1], kDTDB_Chars);
		offset += ckp->keyLength + 1;

		cfp = (HFSCatalogFile *)((UInt8 *)buffer + offset);
		SET_SWAP_BE16(cfp->recordType, kHFSFileRecord);
		SET_SWAP_BE32(cfp->userInfo.fdType, kDTDB_Type);
		SET_SWAP_BE32(cfp->userInfo.fdCreator, kDTDB_Creator);
		SET_SWAP_BE16(cfp->userInfo.fdFlags, kIsInvisible);
		SET_SWAP_BE32(cfp->fileID, kDTDB_FileID);
		SET_SWAP_BE32(cfp->createDate, timeStamp);
		SET_SWAP_BE32(cfp->modifyDate, timeStamp);
		SET_SWAP_BE16(cfp->dataExtents[0].startBlock, gDTDBFork.startBlock);
		SET_SWAP_BE16(cfp->dataExtents[0].blockCount, gDTDBFork.blockCount);
		SET_SWAP_BE32(cfp->dataPhysicalSize, gDTDBFork.physicalSize);
		SET_SWAP_BE32(cfp->dataLogicalSize, gDTDBFork.logicalSize);
		offset += sizeof(HFSCatalogFile);

		SETOFFSET(buffer, nodeSize, offset, 4);

		/*
		 * Add empty "Desktop DF" file...
		 */
		ckp = (HFSCatalogKey *)((UInt8 *)buffer + offset);
		ckp->keyLength			= 1 + 4 + ((kDTDF_Chars + 2) & 0xFE);  /* pad to word */
		SET_SWAP_BE32(ckp->parentID, kHFSRootFolderID);
		ckp->nodeName[0]		= kDTDF_Chars;
		bcopy(kDTDF_Name, &ckp->nodeName[1], kDTDF_Chars);
		offset += ckp->keyLength + 1;

		cfp = (HFSCatalogFile *)((UInt8 *)buffer + offset);
		SET_SWAP_BE16(cfp->recordType, kHFSFileRecord);
		SET_SWAP_BE32(cfp->userInfo.fdType, kDTDF_Type);
		SET_SWAP_BE32(cfp->userInfo.fdCreator, kDTDF_Creator);
		SET_SWAP_BE16(cfp->userInfo.fdFlags, kIsInvisible);
		SET_SWAP_BE32(cfp->fileID, kDTDF_FileID);
		SET_SWAP_BE32(cfp->createDate, timeStamp);
		SET_SWAP_BE32(cfp->modifyDate, timeStamp);
		offset += sizeof(HFSCatalogFile);

		SETOFFSET(buffer, nodeSize, offset, 5);
	}
}


static void
InitSecondCatalogLeaf(const hfsparams_t *dp, void * buffer)
{
	BTNodeDescriptor	*ndp;
	HFSCatalogKey		*ckp;
	HFSCatalogFile		*cfp;
	UInt16			nodeSize;
	SInt16			offset;
	UInt32			timeStamp;

	nodeSize = dp->catalogNodeSize;
	timeStamp = UTCToLocal(dp->createDate);
	bzero(buffer, nodeSize);

	/*
	 * All nodes have a node descriptor...
	 */
	ndp = (BTNodeDescriptor *)buffer;
	SET_SWAP_BE32(ndp->bLink, 2);
	ndp->kind		= kBTLeafNode;
	SET_SWAP_BE16(ndp->numRecords, 3);
	ndp->height		= 1;
	offset = sizeof(BTNodeDescriptor);

	SETOFFSET(buffer, nodeSize, offset, 1);

	/*
	 * Add "Finder" file...
	 */
	ckp = (HFSCatalogKey *)((UInt8 *)buffer + offset);
	ckp->keyLength			= 1 + 4 + ((kFinder_Chars + 2) & 0xFE);  /* pad to word */
	SET_SWAP_BE32(ckp->parentID, kHFSRootFolderID);
	ckp->nodeName[0]		= kFinder_Chars;
	bcopy(kFinder_Name, &ckp->nodeName[1], kFinder_Chars);
	offset += ckp->keyLength + 1;

	cfp = (HFSCatalogFile *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(cfp->recordType, kHFSFileRecord);
	SET_SWAP_BE32(cfp->userInfo.fdType, kFinder_Type);
	SET_SWAP_BE32(cfp->userInfo.fdCreator, kFinder_Creator);
	SET_SWAP_BE16(cfp->userInfo.fdFlags, kIsInvisible + kNameLocked + kHasBeenInited);
	SET_SWAP_BE32(cfp->fileID, kFinder_FileID);
	SET_SWAP_BE32(cfp->createDate, timeStamp);
	SET_SWAP_BE32(cfp->modifyDate, timeStamp);
	offset += sizeof(HFSCatalogFile);

	SETOFFSET(buffer, nodeSize, offset, 2);

	/*
	 * Add "ReadMe" file...
	 */
	ckp = (HFSCatalogKey *)((UInt8 *)buffer + offset);
	ckp->keyLength			= 1 + 4 + ((kReadMe_Chars + 2) & 0xFE);  /* pad to word */
	SET_SWAP_BE32(ckp->parentID, kHFSRootFolderID);
	ckp->nodeName[0]		= kReadMe_Chars;
	bcopy(kReadMe_Name, &ckp->nodeName[1], kReadMe_Chars);
	offset += ckp->keyLength + 1;

	cfp = (HFSCatalogFile *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(cfp->recordType, kHFSFileRecord);
	SET_SWAP_BE32(cfp->userInfo.fdType, kReadMe_Type);
	SET_SWAP_BE32(cfp->userInfo.fdCreator, kReadMe_Creator);
	SET_SWAP_BE16(cfp->userInfo.fdFlags, kIsInvisible);
	SET_SWAP_BE32(cfp->fileID, kReadMe_FileID);
	SET_SWAP_BE32(cfp->createDate, timeStamp);
	SET_SWAP_BE32(cfp->modifyDate, timeStamp);
	SET_SWAP_BE16(cfp->dataExtents[0].startBlock, gReadMeFork.startBlock);
	SET_SWAP_BE16(cfp->dataExtents[0].blockCount, gReadMeFork.blockCount);
	SET_SWAP_BE32(cfp->dataPhysicalSize, gReadMeFork.physicalSize);
	SET_SWAP_BE32(cfp->dataLogicalSize, gReadMeFork.logicalSize);
	offset += sizeof(HFSCatalogFile);

	SETOFFSET(buffer, nodeSize, offset, 3);

	/*
	 * Add "System" file...
	 */
	ckp = (HFSCatalogKey *)((UInt8 *)buffer + offset);
	ckp->keyLength			= 1 + 4 + ((kSystem_Chars + 2) & 0xFE);  /* pad to word */
	SET_SWAP_BE32(ckp->parentID, kHFSRootFolderID);
	ckp->nodeName[0]		= kSystem_Chars;
	bcopy(kSystem_Name, &ckp->nodeName[1], kSystem_Chars);
	offset += ckp->keyLength + 1;

	cfp = (HFSCatalogFile *)((UInt8 *)buffer + offset);
	SET_SWAP_BE16(cfp->recordType, kHFSFileRecord);
	SET_SWAP_BE32(cfp->userInfo.fdType, kSystem_Type);
	SET_SWAP_BE32(cfp->userInfo.fdCreator, kSystem_Creator);
	SET_SWAP_BE16(cfp->userInfo.fdFlags, kIsInvisible + kNameLocked + kHasBeenInited);
	SET_SWAP_BE32(cfp->fileID, kSystem_FileID);
	SET_SWAP_BE32(cfp->createDate, timeStamp);
	SET_SWAP_BE32(cfp->modifyDate, timeStamp);
	SET_SWAP_BE16(cfp->rsrcExtents[0].startBlock, gSystemFork.startBlock);
	SET_SWAP_BE16(cfp->rsrcExtents[0].blockCount, gSystemFork.blockCount);
	SET_SWAP_BE32(cfp->rsrcPhysicalSize, gSystemFork.physicalSize);
	SET_SWAP_BE32(cfp->rsrcLogicalSize, gSystemFork.logicalSize);
	offset += sizeof(HFSCatalogFile);

	SETOFFSET(buffer, nodeSize, offset, 4);
}


static void
InitCatalogRoot_HFS(const hfsparams_t *dp, void * buffer)
{
	BTNodeDescriptor	*ndp;
	HFSCatalogKey		*ckp;
	u_big_int32_t		*prp;	/* pointer record */
	UInt16				nodeSize;
	SInt16				offset;

	nodeSize = dp->catalogNodeSize;
	bzero(buffer, nodeSize);

	/*
	 * All nodes have a node descriptor...
	 */
	ndp = (BTNodeDescriptor *)buffer;
	ndp->kind		= kBTIndexNode;
	SET_SWAP_BE16(ndp->numRecords, 2);
	ndp->height		= 2;
	offset = sizeof(BTNodeDescriptor);

	SETOFFSET(buffer, nodeSize, offset, 1);

	/*
	 * Add root directory index...
	 */
	ckp = (HFSCatalogKey *)((UInt8 *)buffer + offset);
	ckp->keyLength		= kHFSCatalogKeyMaximumLength;
	SET_SWAP_BE32(ckp->parentID, kHFSRootParentID);
	ckp->nodeName[0]	= strlen(dp->volumeName);
	bcopy(dp->volumeName, &ckp->nodeName[1], ckp->nodeName[0]);
	offset += ckp->keyLength + 1;

	prp = (u_big_int32_t *)((UInt8 *)buffer + offset);
	SET_SWAP_BE32(*prp, 2);	/* point to first leaf node */
	offset += sizeof(UInt32);

	SETOFFSET(buffer, nodeSize, offset, 2);

	/*
	 * Add finder file index...
	 */
	ckp = (HFSCatalogKey *)((UInt8 *)buffer + offset);
	ckp->keyLength		= kHFSCatalogKeyMaximumLength;
	SET_SWAP_BE32(ckp->parentID, kHFSRootFolderID);
	ckp->nodeName[0]	= kFinder_Chars;
	bcopy(kFinder_Name, &ckp->nodeName[1], kFinder_Chars);
	offset += ckp->keyLength + 1;

	prp = (u_big_int32_t *)((UInt8 *)buffer + offset);
	SET_SWAP_BE32(*prp, 3);	/* point to last leaf node */
	offset += sizeof(UInt32);

	SETOFFSET(buffer, nodeSize, offset, 3);
}


static void
WriteDesktopDB(const hfsparams_t *dp, const DriveInfo *driveInfo,
               UInt32 startingSector, void *buffer, UInt32 *mapNodes)
{
	BTNodeDescriptor *ndp;
	BTHeaderRec	*bthp;
	UInt8		*bmp;
	UInt32		nodeBitsInHeader;
	UInt32		fileSize;
	UInt32		nodeSize;
	UInt32		temp;
	SInt16		offset;
	UInt8		*keyDiscP;

	*mapNodes = 0;
	fileSize = gDTDBFork.logicalSize;
	nodeSize = kHFSNodeSize;

	bzero(buffer, nodeSize);

	/* FILL IN THE NODE DESCRIPTOR:  */
	ndp = (BTNodeDescriptor *)buffer;
	ndp->kind		= kBTHeaderNode;
	SET_SWAP_BE16(ndp->numRecords, 3);
	offset = sizeof(BTNodeDescriptor);

	SETOFFSET(buffer, nodeSize, offset, 1);

	/* FILL IN THE HEADER RECORD:  */
	bthp = (BTHeaderRec *)((UInt8 *)buffer + offset);
    //	bthp->treeDepth		= 0;
    //	bthp->rootNode		= 0;
    //	bthp->firstLeafNode	= 0;
    //	bthp->lastLeafNode	= 0;
    //	bthp->leafRecords	= 0;
	SET_SWAP_BE16(bthp->nodeSize, nodeSize);
	SET_SWAP_BE16(bthp->maxKeyLength, 37);
	SET_SWAP_BE32(bthp->totalNodes, fileSize / nodeSize);
	SET_SWAP_BE32(bthp->freeNodes, SWAP_BE32 (bthp->totalNodes) - 1);  /* header */
	SET_SWAP_BE32(bthp->clumpSize, fileSize);
	bthp->btreeType		= 0xFF;
	offset += sizeof(BTHeaderRec);

	SETOFFSET(buffer, nodeSize, offset, 2);
	
	keyDiscP = (UInt8 *)((UInt8 *)buffer + offset);
	*keyDiscP++ = 2;			/* length of descriptor */
	*keyDiscP++ = KD_USEPROC;	/* always uses a compare proc */
	*keyDiscP++ = 1;			/* just one of them */
	offset += kBTreeHeaderUserBytes;

	SETOFFSET(buffer, nodeSize, offset, 3);

	/* FIGURE OUT HOW MANY MAP NODES (IF ANY):  */
	nodeBitsInHeader = 8 * (nodeSize
					- sizeof(BTNodeDescriptor)
					- sizeof(BTHeaderRec)
					- kBTreeHeaderUserBytes
					- (4 * sizeof(SInt16)) );

	if (SWAP_BE32 (bthp->totalNodes) > nodeBitsInHeader) {
		UInt32	nodeBitsInMapNode;
		
		SET_SWAP_BE32(ndp->fLink, SWAP_BE32 (bthp->lastLeafNode) + 1);
		nodeBitsInMapNode = 8 * (nodeSize
						- sizeof(BTNodeDescriptor)
						- (2 * sizeof(SInt16))
						- 2 );
		*mapNodes = (SWAP_BE32 (bthp->totalNodes) - nodeBitsInHeader +
		            (nodeBitsInMapNode - 1)) / nodeBitsInMapNode;
		SET_SWAP_BE32(bthp->freeNodes, SWAP_BE32 (bthp->freeNodes) - *mapNodes);
	}

	/* 
	 * FILL IN THE MAP RECORD, MARKING NODES THAT ARE IN USE.
	 * Note - worst case (32MB alloc blk) will have only 18 nodes in use.
	 */
	bmp = ((UInt8 *)buffer + offset);
	temp = SWAP_BE32 (bthp->totalNodes) - SWAP_BE32 (bthp->freeNodes);

	/* Working a byte at a time is endian safe */
	while (temp >= 8) { *bmp = 0xFF; temp -= 8; bmp++; }
	*bmp = ~(0xFF >> temp);
	offset += nodeBitsInHeader/8;

	SETOFFSET(buffer, nodeSize, offset, 4);

	WriteBuffer(driveInfo, startingSector, kHFSNodeSize, buffer);
}


static void
WriteSystemFile(const DriveInfo *dip, UInt32 startingSector, UInt32 *filesize)
{
	int fd;
	ssize_t	datasize, writesize;
	UInt8 *buf;
	struct stat stbuf;
	
	if (stat(HFS_BOOT_DATA, &stbuf) < 0)
		err(1, "stat %s", HFS_BOOT_DATA);
	
	datasize = stbuf.st_size;
	writesize = ROUNDUP(datasize, dip->sectorSize);

	if (datasize > (64 * 1024))
		errx(1, "hfsbootdata file too big.");

	if ((buf = malloc(writesize)) == NULL)
		exit(1);
	
	if ((fd = open(HFS_BOOT_DATA, O_RDONLY, 0)) < 0)
		err(1, "open %s", HFS_BOOT_DATA);
	
	if (read(fd, buf, datasize) != datasize) {
		if (errno)
			err(1, "read %s", HFS_BOOT_DATA);
		else
			errx(1, "problems reading %s", HFS_BOOT_DATA);
	}

	if (writesize > datasize)
		bzero(buf + datasize, writesize - datasize);

	WriteBuffer(dip, startingSector, writesize, buf);
	
	close(fd);
	
	free(buf);
	
	*filesize = datasize;
}


static void
WriteReadMeFile(const DriveInfo *dip, UInt32 startingSector, UInt32 *filesize)
{
	ssize_t	datasize, writesize;
	UInt8 *buf;
	
	datasize = sizeof(hfswrap_readme);
	writesize = ROUNDUP(datasize, dip->sectorSize);

	if ((buf = malloc(writesize)) == NULL)
		exit(1);
	
	bcopy(hfswrap_readme, buf, datasize);
	if (writesize > datasize)
		bzero(buf + datasize, writesize - datasize);
	
	WriteBuffer(dip, startingSector, writesize, buf);
	
	*filesize = datasize;
}


/*
 * WriteMapNodes
 *	
 * Initializes a B-tree map node and writes it out to disk.
 */
static void
WriteMapNodes(const DriveInfo *driveInfo, UInt32 diskStart, UInt32 firstMapNode,
	UInt32 mapNodes, UInt16 btNodeSize, void *buffer)
{
	UInt32	sectorsPerNode;
	UInt32	mapRecordBytes;
	UInt16	i;
	BTNodeDescriptor *nd = (BTNodeDescriptor *)buffer;

	bzero(buffer, btNodeSize);

	nd->kind		= kBTMapNode;
	SET_SWAP_BE16(nd->numRecords, 1);
	
	/* note: must belong word aligned (hence the extra -2) */
	mapRecordBytes = btNodeSize - sizeof(BTNodeDescriptor) - 2*sizeof(SInt16) - 2;	

	SETOFFSET(buffer, btNodeSize, sizeof(BTNodeDescriptor), 1);
	SETOFFSET(buffer, btNodeSize, sizeof(BTNodeDescriptor) + mapRecordBytes, 2);
	
	sectorsPerNode = btNodeSize/kBytesPerSector;
	
	/*
	 * Note - worst case (32MB alloc blk) will have
	 * only 18 map nodes. So don't bother optimizing
	 * this section to do multiblock writes!
	 */
	for (i = 0; i < mapNodes; i++) {
		if ((i + 1) < mapNodes)
			SET_SWAP_BE32(nd->fLink, ++firstMapNode);  /* point to next map node */
		else
			SET_SWAP_BE32(nd->fLink, 0);  /* this is the last map node */

		WriteBuffer(driveInfo, diskStart, btNodeSize, buffer);
			
		diskStart += sectorsPerNode;
	}
}

/*
 * ClearDisk
 *
 * Clear out consecutive sectors on the disk.
 *
 */
static void
ClearDisk(const DriveInfo *driveInfo, UInt64 startingSector, UInt32 numberOfSectors)
{
	UInt32 bufferSize;
	UInt32 bufferSizeInSectors;
	void *tempBuffer = NULL;


	if ( numberOfSectors > driveInfo->sectorsPerIO )
		bufferSizeInSectors = driveInfo->sectorsPerIO;
	else
		bufferSizeInSectors = numberOfSectors;

	bufferSize = bufferSizeInSectors << kLog2SectorSize;

	tempBuffer = valloc((size_t)bufferSize);
	if (tempBuffer == NULL)
		exit(1);

	bzero(tempBuffer, bufferSize);

	while (numberOfSectors > 0) {
		WriteBuffer(driveInfo, startingSector, bufferSize, tempBuffer);
	
		startingSector += bufferSizeInSectors;	
		numberOfSectors -= bufferSizeInSectors;
		
		/* is remainder less than size of buffer? */
		if (numberOfSectors < bufferSizeInSectors) {
			bufferSizeInSectors = numberOfSectors;
			bufferSize = bufferSizeInSectors << kLog2SectorSize;
		}
	}

	if (tempBuffer)
		free(tempBuffer);
}


static void
WriteBuffer(const DriveInfo *driveInfo, UInt64 startingSector, UInt32 byteCount,
	const void *buffer)
{
	off_t sector;

	if ((byteCount % driveInfo->sectorSize) != 0)
		errx(1, "WriteBuffer: byte count %u is not sector size multiple", byteCount);

	sector = driveInfo->sectorOffset + startingSector;

	if (lseek(driveInfo->fd, sector * driveInfo->sectorSize, SEEK_SET) < 0)
		err(1, "seek (sector %qd)", sector);

	if (write(driveInfo->fd, buffer, byteCount) != byteCount)
		err(1, "write (sector %qd, %u bytes)", sector, byteCount);
}


static UInt32 Largest( UInt32 a, UInt32 b, UInt32 c, UInt32 d )
{
	/* a := max(a,b) */
	if (a < b)
		a = b;
	/* c := max(c,d) */
	if (c < d)
		c = d;
	
	/* return max(a,c) */
	if (a > c)
		return a;
	else
		return c;
}


/*
 * MarkBitInAllocationBuffer
 * 	
 * Given a buffer and allocation block, will mark off the corresponding
 * bitmap bit, and return the sector number the block belongs in.
 */
static void MarkBitInAllocationBuffer( HFSPlusVolumeHeader *header,
	UInt32 allocationBlock, void* sectorBuffer, UInt32 *sector )
{

	UInt8 *byteP;
	UInt8 mask;
	UInt32 sectorsPerBlock;
	UInt16 bitInSector = allocationBlock % kBitsPerSector;
	UInt16 bitPosition = allocationBlock % 8;
	
	sectorsPerBlock = SWAP_BE32(header->blockSize) / kBytesPerSector;

	*sector = (SWAP_BE32(header->allocationFile.extents[0].startBlock) * sectorsPerBlock) +
		  (allocationBlock / kBitsPerSector);
	
	byteP = (UInt8 *)sectorBuffer + (bitInSector >> 3);
	mask = ( 0x80 >> bitPosition );
	*byteP |= mask;
}


/*
 * UTCToLocal - convert from Mac OS GMT time to Mac OS local time
 */
static UInt32 UTCToLocal(UInt32 utcTime)
{
	UInt32 localTime = utcTime;
	struct timezone timeZone;
	struct timeval	timeVal;
	
	if (localTime != 0) {

                /* HFS volumes need timezone info to convert local to GMT */
                (void)gettimeofday( &timeVal, &timeZone );


		localTime -= (timeZone.tz_minuteswest * 60);
		if (timeZone.tz_dsttime)
			localTime += 3600;
	}

        return (localTime);
}


static UInt32
DivideAndRoundUp(UInt32 numerator, UInt32 denominator)
{
	UInt32	quotient;
	
	quotient = numerator / denominator;
	if (quotient * denominator != numerator)
		quotient++;
	
	return quotient;
}


#define __kCFUserEncodingFileName ("/.CFUserTextEncoding")

static UInt32
GetDefaultEncoding()
{
    struct passwd *passwdp;

    if ((passwdp = getpwuid(0))) { // root account
        char buffer[MAXPATHLEN + 1];
        int fd;

        strcpy(buffer, passwdp->pw_dir);
        strcat(buffer, __kCFUserEncodingFileName);

        if ((fd = open(buffer, O_RDONLY, 0)) > 0) {
            size_t readSize;

            readSize = read(fd, buffer, MAXPATHLEN);
            buffer[(readSize < 0 ? 0 : readSize)] = '\0';
            close(fd);
            return strtol(buffer, NULL, 0);
        }
    }
    return 0;
}


static int
ConvertUTF8toUnicode(const UInt8* source, UInt32 bufsize, UniChar* unibuf,
	u_big_int16_t *charcount)
{
	UInt8 byte;
	UniChar* target;
	UniChar* targetEnd;

	SET_SWAP_BE16(*charcount, 0);
	target = unibuf;
	targetEnd = (UniChar *)((UInt8 *)unibuf + bufsize);

	while ((byte = *source++)) {
		
		/* check for single-byte ascii */
		if (byte < 128) {
			if (byte == ':')	/* ':' is mapped to '/' */
				byte = '/';

			SET_SWAP_BE16(*target++, byte);
		} else {
			UInt16 ch;
			UInt8 seq = (byte >> 4);

			switch (seq) {
			case 0xc: /* double-byte sequence (1100 and 1101) */
			case 0xd:
				ch = (byte & 0x1F) << 6;  /* get 5 bits */
				if (((byte = *source++) >> 6) != 2)
					return (EINVAL);
				break;

			case 0xe: /* triple-byte sequence (1110) */
				ch = (byte & 0x0F) << 6;  /* get 4 bits */
				if (((byte = *source++) >> 6) != 2)
					return (EINVAL);
				ch += (byte & 0x3F); ch <<= 6;  /* get 6 bits */
				if (((byte = *source++) >> 6) != 2)
					return (EINVAL);
				break;

			default:
				return (EINVAL);  /* malformed sequence */
			}

			ch += (byte & 0x3F);  /* get last 6 bits */

			if (target >= targetEnd)
				return (ENOBUFS);

			SET_SWAP_BE16(*target++, ch);
		}
	}

	SET_SWAP_BE16(*charcount, target - unibuf);

	return (0);
}
